"""
Tests for `worker` module.
"""

import pytest

from omnic import singletons
from omnic.worker.aioworker import AioWorker
from omnic.worker.base import BaseWorker
from omnic.worker.enums import Task
from omnic.worker.testing import ForegroundWorker, autodrain_worker

from .testing_utils import MockAioQueue, MockWorker


class WorkerTestBase:
    @classmethod
    def setup_class(cls):
        class FakeSettings:
            LOGGING = None
        singletons.settings.use_settings(FakeSettings)

    @classmethod
    def teardown_class(cls):
        singletons.settings.use_previous_settings()


class TestBaseWorker(WorkerTestBase):
    @pytest.mark.asyncio
    async def test_run_func(self):
        worker = MockWorker()

        def fake_func(*args):
            worker.called_args = args
        worker.fake_next = (Task.FUNC, (fake_func, 1, 2, 3))
        res = await worker.run()
        assert res is None
        assert worker.called_args == (1, 2, 3)

    def test_get_method(self):
        worker = MockWorker()
        assert worker._get_method(Task.FUNC) == worker.run_func
        assert worker._get_method(Task.DOWNLOAD) == worker.run_download
        assert worker._get_method(Task.CONVERT) == worker.run_convert


class TestAsyncioWorker(WorkerTestBase):
    @pytest.mark.asyncio
    async def test_run_once(self):
        worker = AioWorker(MockAioQueue())
        worker.called = 0

        def fake_func(*args):
            called = worker.called
            assert args == (1 + called, 2 + called, 3 + called)
            worker.called += 1
            if worker.called > 3:
                worker.running = False
        await worker.enqueue(Task.FUNC, (fake_func, 1, 2, 3))
        await worker.enqueue(Task.FUNC, (fake_func, 2, 3, 4))
        await worker.enqueue(Task.FUNC, (fake_func, 3, 4, 5))
        await worker.enqueue(Task.FUNC, (fake_func, 4, 5, 6))
        await worker.enqueue(Task.FUNC, (fake_func, 5, 6, 7))
        res = await worker.run()
        assert res is None
        assert worker.called == 4


class TestForegroundWorker(WorkerTestBase):
    @pytest.mark.asyncio
    async def test_run(self):
        worker = ForegroundWorker()
        worker.called = 0

        def fake_func(*args):
            called = worker.called
            assert args == (1 + called, 2 + called, 3 + called)
            worker.called += 1
        await worker.enqueue(Task.FUNC, (fake_func, 1, 2, 3))
        await worker.enqueue(Task.FUNC, (fake_func, 2, 3, 4))
        await worker.enqueue(Task.FUNC, (fake_func, 3, 4, 5))
        assert worker.called == 3

    @pytest.mark.asyncio
    async def test_worker_decorator(self):
        self.called = 0

        def fake_func(*args):
            assert args == (1 + self.called, 2 + self.called, 3 + self.called)
            self.called += 1

        with autodrain_worker() as worker:
            assert worker
            assert isinstance(worker, BaseWorker)
            await singletons.workers.async_enqueue_sync(fake_func, 1, 2, 3)
            await singletons.workers.async_enqueue_sync(fake_func, 2, 3, 4)
            await singletons.workers.async_enqueue_sync(fake_func, 3, 4, 5)
        assert self.called == 3



"""
Tests for `utils` module.
"""
import tempfile
from os.path import islink, join

import pytest

from omnic.utils import filesystem, graph, iters

from .testing_utils import clear_tmp_files, gen_tmp_files


class TestDirectedGraph:
    #  ,-> F ,-> E
    # A -> B  -> C
    #        '-> D
    def _simple_tree(self):
        self.dg = graph.DirectedGraph()
        self.dg.add_edge('A', 'B', 1)
        self.dg.add_edge('A', 'F', 1)
        self.dg.add_edge('B', 'E', 1)
        self.dg.add_edge('B', 'C', 1)
        self.dg.add_edge('B', 'D', 1)

    def _add_preferred_path(self):
        self.dg.add_preferred_path('A', 'Z', 'B')

    #  ,-> F - - - -.
    # '    v         v
    # A -> B -> G -> C
    def _multi_pathed_graph(self):
        self.dg = graph.DirectedGraph()
        self.dg.add_edge('A', 'B', 1)
        self.dg.add_edge('B', 'G', 1)
        self.dg.add_edge('G', 'C', 1)
        self.dg.add_edge('A', 'F', 1)
        self.dg.add_edge('F', 'C', 1)
        self.dg.add_edge('F', 'B', 1)

    #  5-> F - - - -.
    # '    v         v
    # A -> B -> G -> C
    def _multi_pathed_graph_weighted(self):
        self.dg = graph.DirectedGraph()
        self.dg.add_edge('A', 'B', 1)
        self.dg.add_edge('B', 'G', 1)
        self.dg.add_edge('G', 'C', 1)
        self.dg.add_edge('A', 'F', 5)
        self.dg.add_edge('F', 'C', 1)
        self.dg.add_edge('F', 'B', 1)

    #  STL  .   MOV .   GIF .
    #  OBJ   -> AVI  -> JPG  -> thumb.png
    #  MESH '   MP4 '   PNG '
    #      \\\         .^
    #       ''---------
    #  MP3 .
    #  WAV  -> cleaned.ogg
    #  OGG '

    def _realistic_edges(self):
        self.dg = graph.DirectedGraph()
        self.dg.add_edge('MOV', 'JPG')
        self.dg.add_edge('AVI', 'JPG')
        self.dg.add_edge('MP4', 'JPG')
        self.dg.add_edge('JPG', 'thumb.png')
        self.dg.add_edge('PNG', 'thumb.png')
        self.dg.add_edge('GIF', 'thumb.png')
        self.dg.add_edge('STL', 'AVI')
        self.dg.add_edge('OBJ', 'AVI')
        self.dg.add_edge('MESH', 'AVI')
        self.dg.add_edge('STL', 'PNG')
        self.dg.add_edge('OBJ', 'PNG')
        self.dg.add_edge('MESH', 'PNG')
        self.dg.add_edge('MP3', 'cleaned.ogg')
        self.dg.add_edge('WAV', 'cleaned.ogg')
        self.dg.add_edge('OGG', 'cleaned.ogg')

    def test_simple_routes(self):
        self._simple_tree()
        path = self.dg.shortest_path('A', 'B')
        assert path == ('A', 'B')
        path = self.dg.shortest_path('A', 'C')
        assert path == ('A', 'B', 'C')
        path = self.dg.shortest_path('A', 'E')
        assert path == ('A', 'B', 'E')
        path = self.dg.shortest_path('A', 'D')
        assert path == ('A', 'B', 'D')
        path = self.dg.shortest_path('B', 'C')
        assert path == ('B', 'C')

    def test_special_paths(self):
        self._simple_tree()
        self._add_preferred_path()
        path = self.dg.shortest_path('A', 'B')
        assert path == ('A', 'Z', 'B')
        path = self.dg.shortest_path('A', 'C')
        assert path == ('A', 'B', 'C')

    def test_raises_on_invalid_path(self):
        self._simple_tree()
        with pytest.raises(graph.DirectedGraph.NoPath):
            self.dg.shortest_path('B', 'A')

    def test_shortest_route(self):
        self._multi_pathed_graph()
        path = self.dg.shortest_path('A', 'B')
        assert path == ('A', 'B')
        path = self.dg.shortest_path('F', 'B')
        assert path == ('F', 'B')
        path = self.dg.shortest_path('F', 'G')
        assert path == ('F', 'B', 'G')
        path = self.dg.shortest_path('F', 'C')
        assert path == ('F', 'C')
        path = self.dg.shortest_path('A', 'G')
        assert path == ('A', 'B', 'G')
        path = self.dg.shortest_path('A', 'C')
        assert path == ('A', 'F', 'C')

    def test_weighted_shortest_route(self):
        self._multi_pathed_graph_weighted()
        path = self.dg.shortest_path('A', 'B')
        assert path == ('A', 'B')
        path = self.dg.shortest_path('F', 'B')
        assert path == ('F', 'B')
        path = self.dg.shortest_path('F', 'G')
        assert path == ('F', 'B', 'G')
        path = self.dg.shortest_path('F', 'C')
        assert path == ('F', 'C')
        path = self.dg.shortest_path('A', 'G')
        assert path == ('A', 'B', 'G')
        path = self.dg.shortest_path('A', 'C')
        assert path == ('A', 'B', 'G', 'C')  # Avoid A->F route

    def test_realistic_routing(self):
        self._realistic_edges()
        path = self.dg.shortest_path('STL', 'thumb.png')
        assert path == ('STL', 'PNG', 'thumb.png')
        path = self.dg.shortest_path('MESH', 'JPG')
        assert path == ('MESH', 'AVI', 'JPG')
        path = self.dg.shortest_path('MP3', 'cleaned.ogg')
        assert path == ('MP3', 'cleaned.ogg')
        with pytest.raises(graph.DirectedGraph.NoPath):
            self.dg.shortest_path('MP3', 'thumb.png')


class TestIterUtils:
    def test_pair_looper(self):
        pair_looper = iters.pair_looper
        assert list(pair_looper([1, 2, 3])) == [(1, 2), (2, 3)]
        assert list(pair_looper([1, 2, 3, 4])) == [(1, 2), (2, 3), (3, 4)]
        assert list(pair_looper([1])) == []
        assert list(pair_looper([])) == []

    def test_first_last_iterator(self):
        first_last_iterator = iters.first_last_iterator
        for first, last, i in first_last_iterator([1, 2, 3]):
            assert i in (1, 2, 3)
            if i == 1:
                assert first and not last
            if i == 2:
                assert not first and not last
            if i == 3:
                assert not first and last

        for first, last, i in first_last_iterator([1]):
            assert i == 1
            assert first and last

    def test_group_by(self):
        group_by = iters.group_by
        assert list(group_by('asdf', 2)) == ['as', 'df']


class TestFilesystemUtils:
    FILES = [
        'testfile',
        'config/thing/thing.xml',
    ]

    def setup_method(self, method):
        self.dir = tempfile.mkdtemp(prefix='tmp_omnic_test_')
        self.out = tempfile.mkdtemp(prefix='tmp_omnic_test_out_')
        gen_tmp_files(self.dir, self.FILES)
        self.results = set([
            (join(self.dir, 'testfile'),
                join(self.out, 'testfile')),
            (join(self.dir, 'config/thing/thing.xml'),
                join(self.out, 'config/thing/thing.xml')),
        ])

    def teardown_method(self, method):
        clear_tmp_files(self.dir, self.FILES)
        clear_tmp_files(self.out, self.FILES)

    def test_directory_walk(self):
        results = list(filesystem.directory_walk(self.dir, self.out))
        results_set = set(results)
        assert len(results_set) == len(results)  # ensure no dupes
        assert results_set == self.results

    def test_recursive_symlinks(self):
        filesystem.recursive_symlink_dirs(self.dir, self.out)

        # Now use walk to check that we did it successfully
        results = set(filesystem.directory_walk(self.out, self.out))
        assert len(results) == len(self.results)  # ensure right number
        for path, _ in results:
            assert islink(path)

    def test_recursive_hardlinks(self):
        filesystem.recursive_hardlink_dirs(self.dir, self.out)

        # Now use walk to check that we did it successfully
        results = set(filesystem.directory_walk(self.out, self.out))
        assert len(results) == len(self.results)  # ensure right number
        for path, _ in results:
            assert not islink(path)  # ensure hardlinks



import os
from base64 import b64decode
from os.path import exists, join

from omnic.types.detectors import Detector
from omnic.worker.base import BaseWorker


class Magic:
    JPEG = bytes([0xff, 0xd8, 0xff, 0xe0])
    PNG = bytes([0x89, 0x50, 0x4e, 0x47])
    PNG_PIXEL_B64 = 'iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR42mNkYAAAAAYAAjCB0C8AAAAASUVORK5CYII='
    PNG_PIXEL_B = b64decode(PNG_PIXEL_B64)


class MockWorker(BaseWorker):
    '''
    Test worker to test abstract Worker base class
    '''

    def __init__(self):
        super().__init__()
        self.fake_next = None
        self._running = True
        self.called_args = None

    @property
    def running(self):
        if not self._running:
            return False
        self._running = False
        return True

    async def get_next(self):
        return self.fake_next

    async def check_download(self, foreign_resource):
        self.check_download_was_called = True
        return True

    async def check_convert(self, converter, in_r, out_r):
        self.check_convert_was_called = True
        return True


class MockAioQueue(list):
    async def put(self, item):
        self.append(item)

    async def get(self):
        return self.pop(0)


class DummyDetector(Detector):
    pass


class AgreeableDetector(Detector):
    def can_detect(self, path):
        return True

    def detect(self, path):
        return 'something'


def rm_tmp_file(path):
    # Utility function for completely removing a temp file and any superfluous
    # dirs used to contain it
    try:
        os.remove(path)
    except OSError:
        pass
    try:
        os.removedirs(os.path.dirname(path))
    except OSError:
        pass


def rm_tmp_files(*paths, prefixes=[]):
    # For a group of paths that could be in multiple locations
    for path in paths:
        if prefixes:
            for prefix in prefixes:
                rm_tmp_file(os.path.join(prefix, path))
        else:
            rm_tmp_file(path)


# Some testing utilities borrowed from another python package (stowage)
# TODO: refactor these into above
def write_tmp_file(path):
    try:
        os.makedirs(os.path.dirname(path))
    except OSError:
        pass
    with open(path, 'w+') as f:
        f.write('%s contents' % path)


def gen_tmp_files(root, files):
    for fn in files:
        write_tmp_file(join(root, fn))


def clear_tmp_files(root, files):
    for fn in files:
        path = join(root, fn)
        if exists(path):
            os.remove(path)
        try:
            os.removedirs(os.path.dirname(path))
        except OSError:
            pass



from omnic import singletons


class TestSettingsBuiltinDefaults:
    '''
    High-level sanity tests for built-ins
    '''
    @classmethod
    def setup_class(cls):
        # Ensure no custom settings, except ignoring the system check
        class config:
            CONVERSION_SYSTEM_CHECK = False
        singletons.settings.use_settings(config)

    @classmethod
    def teardown_class(cls):
        # Ensure no custom settings
        singletons.settings.use_previous_settings()

    def test_default_settings_exist(self):
        s = singletons.settings
        assert hasattr(s.default_settings_module, 'CONVERSION_SYSTEM_CHECK')
        assert s.default_settings_module.CONVERSION_SYSTEM_CHECK
        assert hasattr(s, 'PATH_PREFIX') and s.PATH_PREFIX
        assert hasattr(s, 'PATH_GROUPING') and s.PATH_GROUPING
        assert hasattr(s, 'WEB_SERVER') and s.WEB_SERVER
        assert hasattr(s, 'EVENT_LOOP') and s.EVENT_LOOP

    def test_default_modules_exist(self):
        s = singletons.settings
        assert hasattr(s, 'CONVERTERS')
        assert len(s.CONVERTERS) > 7  # shouldn't drop below this
        assert hasattr(s, 'SERVICES')
        assert len(s.SERVICES) > 3  # shouldn't drop below this
        assert hasattr(s, 'PLACEHOLDERS')
        assert hasattr(s, 'ALLOWED_LOCATIONS')
        assert len(s.ALLOWED_LOCATIONS) > 1  # at least localhost
        assert hasattr(s, 'LOGGING')
        assert s.LOGGING.keys()

    def test_default_settings_are_importable(self):
        s = singletons.settings
        assert len(s.load_all('PLACEHOLDERS')) > 1
        assert len(s.load_all('VIEWERS')) > 2
        assert len(s.load_all('AUTOLOAD')) > 3
        assert len(s.load_all('CONVERTERS')) > 7
        assert len(s.load_all('SERVICES')) > 3
        assert s.load('WORKER')

    def test_singletons(self):
        cg = singletons.converter_graph
        assert len(cg.converters) > 7
        v = singletons.viewers
        assert len(v.viewers) > 2

    def test_certain_viewers_exist(self):
        viewers_list = singletons.viewers.viewers
        viewer_types = sum([viewer.types for viewer in viewers_list], [])
        assert 'PDF' in viewer_types
        assert 'STL' in viewer_types



"""
Tests for `responses` module.
"""


from omnic.responses import placeholder
from omnic.types.typestring import TypeString

from .testing_utils import Magic


class ExampleBytesPlaceholder(placeholder.BytesPlaceholder):
    types = [
        'PNG',
        'JPG',
    ]
    content_type = 'image/png'
    bytes = Magic.PNG_PIXEL_B


class ExampleCustomPlaceholder(placeholder.Placeholder):
    types = [
        'JPEG',
        'JPG',
    ]
    content_type = 'image/jpeg'

    async def stream_response(self, response):
        response.write(Magic.JPEG)


class ExampleWildcardPlaceholder(placeholder.Placeholder):
    types = all
    content_type = 'image/png'

    async def stream_response(self, response):
        response.write(Magic.PNG)


class MockConfig:
    PLACEHOLDERS = [
        ExampleBytesPlaceholder,
        ExampleCustomPlaceholder,
        ExampleWildcardPlaceholder,
    ]


class TestResponsePlaceholders:
    @classmethod
    def setup_class(cls):
        cls.phs = placeholder.PlaceholderSelector(MockConfig)

    def test_get_placeholder(self):
        ph = self.phs.get_placeholder(TypeString('PNG'))
        assert isinstance(ph, ExampleBytesPlaceholder)
        ph = self.phs.get_placeholder(TypeString('JPEG'))
        assert isinstance(ph, ExampleCustomPlaceholder)
        ph = self.phs.get_placeholder(TypeString('nonexistent'))
        assert isinstance(ph, ExampleWildcardPlaceholder)

        # Temporarily disable
        ExampleWildcardPlaceholder.types = []
        ph = self.phs.get_placeholder(TypeString('nonexistent'))
        assert ph is None
        ExampleWildcardPlaceholder.types = all  # restore

    def test_placeholder_stream_response(self):
        class mock_response:
            @staticmethod
            def write(data):
                mock_response.data = data

            @staticmethod
            def stream(streamer=None, content_type=None):
                mock_response.ct = content_type
                # For now, too hard to test this bit
                #coro = streamer(mock_response)
                #assert iscoroutine(coro)
                # yield from coro(mock_response)

        self.phs.stream_response(TypeString('PNG'), mock_response)
        assert mock_response.ct == 'image/png'
        #assert mock_response.data == Magic.PNG_PIXEL_B

        self.phs.stream_response(TypeString('JPEG'), mock_response)
        assert mock_response.ct == 'image/jpeg'
        #assert mock_response.data == Magic.JPEG

        self.phs.stream_response(TypeString('lol'), mock_response)
        assert mock_response.ct == 'image/png'
        #assert mock_response.data == Magic.PNG
